@@ -38,13 +38,28 @@ appropriate action can be taken.
 
 =cut
 
-use 5.00503;
+use 5.006;
 use strict;
+use warnings;
 
-$Mail::DeliveryStatus::BounceParser::VERSION = '1.515';
+our $VERSION = '1.525';
+$VERSION = eval $VERSION;
 
 use MIME::Parser;
 use Mail::DeliveryStatus::Report;
+use vars qw($EMAIL_ADDR_REGEX);
+
+$EMAIL_ADDR_REGEX = qr{
+# Avoid using something like Email::Valid
+# Full rfc(2)822 compliance isn't exactly what we want, and this seems to work
+# for most real world cases
+(?:<|^|\s)            # Space, or the start of a string
+([^\s\/<]+            # some non-space, non-/ characters; none are <
+\@                    # at sign (duh)
+(?:[-\w]+\.)+[-\w]+)  # word characters or hypens organized into
+                      # at least two dot-separated words
+(?:$|\s|>)            # then the end
+}sx;
 
 my $Not_An_Error = qr/
     \b delayed \b
@@ -53,23 +68,22 @@ my $Not_An_Error = qr/
   | Your \s message .{0,100} was \s delivered \s to \s the \s following \s recipient
 /six;
 
-my $Really_An_Error = qr/this is a permanent error/i;
+# added "permanent fatal errors" - fix for bug #41874
+my $Really_An_Error = qr/this is a permanent error|permanent fatal errors/i;
 
 my $Returned_Message_Below = qr/(
     (?:original|returned) \s message \s (?:follows|below)
   | (?: this \s is \s a \s copy \s of
       | below \s this \s line \s is \s a \s copy
-    ) .{0,100} \s message
+    ) .{0,100} \s message\.?
   | message \s header \s follows
   | ^ (?:return-path|received|from):
-)/sixm;
+)\s+/sixm;
 
 my @Preprocessors = qw(
-  p_ms
   p_ims
-  p_compuserve
   p_aol_senderblock
-  p_novell_groupwise_5_2
+  p_novell_groupwise
   p_plain_smtp_transcript
   p_xdelivery_status
 );
@@ -83,7 +97,7 @@ used as a logging callback.
 
 NON-BOUNCES.  If the message is recognizably a vacation autoresponse, or is a
 report of a transient nonfatal error, or a spam or virus autoresponse, you'll
-still get back a C<$bounce>, but its C<<$bounce->is_bounce()>> will return
+still get back a C<$bounce>, but its C<< $bounce->is_bounce() >> will return
 false.
 
 It is possible that some bounces are not really bounces; such as
@@ -99,8 +113,10 @@ sub parse {
   # my $bounce = Mail::DeliveryStatus::BounceParser->new( \*STDIN | $fh |
   # "entire\nmessage" | ["array","of","lines"] );
 
-  my $parser = new MIME::Parser;
+  my $parser = MIME::Parser->new;
      $parser->output_to_core(1);
+     $parser->decode_headers(1);
+
   my $message;
 
   if (not $data) {
@@ -120,13 +136,14 @@ sub parse {
     log       => $arg->{log},
     parser    => $parser,
     orig_message_id => undef,
+    prefer_final_recipient => $arg->{prefer_final_recipient},
   }, $class;
 
   $self->log(
     "received message with type "
-    . $message->effective_type
+    . (defined($message->effective_type) ? $message->effective_type : "undef")
     . ", subject "
-    . $message->head->get("subject")
+    . (defined($message->head->get("subject")) ? $message->head->get("subject") : "CAN'T GET SUBJECT")
   );
 
   # before we even start to analyze the bounce, we recognize certain special
@@ -143,7 +160,7 @@ sub parse {
     "now the message is type "
     . $message->effective_type
     . ", subject "
-    . $message->head->get("subject")
+    . (defined($message->head->get("subject")) ? $message->head->get("subject") : "CAN'T GET SUBJECT")
   );
 
   my $first_part = _first_non_multi_part($message);
@@ -158,6 +175,22 @@ sub parse {
     return $self;
   }
 
+  {
+    last unless ($message->head->get("X-Bluebottle-Request") and $first_part->stringify_body =~ /This account is protected by Bluebottle/);
+    $self->log("looks like a challenge/response autoresponse; ignoring.");
+    $self->{type} = "Challenge / Response system autoreply";
+    $self->{is_bounce} = 0;
+    return $self;
+  }
+
+  {
+    last unless $first_part->stringify_body =~ /Your server requires confirmation/;
+    $self->log("Looks like a challenge/response autoresponse; ignoring.");
+    $self->{type} = "Challenge / Response system autoreply";
+    $self->{is_bounce} = 0;
+    return $self;
+  }
+
   # we'll deem autoreplies to be usually less than a certain size.
 
   # Some vacation autoreplies are (sigh) multipart/mixed, with an additional
@@ -173,13 +206,40 @@ sub parse {
     last if !$first_part || $first_part->effective_type ne 'text/plain';
     my $string = $first_part->as_string;
     last if length($string) > 3000;
-    last if $string !~ /auto.{0,20}reply|vacation|(out|away|on holiday).*office/i;
+    # added return receipt (fix for bug #41870)
+    last if $string !~ /auto.{0,20}(reply|response)|return receipt|vacation|(out|away|on holiday).*office/i;
     $self->log("looks like a vacation autoreply, ignoring.");
     $self->{type} = "vacation autoreply";
     $self->{is_bounce} = 0;
     return $self;
   }
 
+  # vacation autoreply tagged in the subject
+  {
+    last if $message->effective_type eq 'multipart/report';
+    last if !$first_part || $first_part->effective_type ne 'text/plain';
+    my $subject = $message->head->get('Subject');
+    last if !defined($subject);
+    last if $subject !~ /^AUTO/;
+    last if $subject !~ /is out of the office/;
+    $self->log("looks like a vacation autoreply, ignoring.");
+    $self->{type} = "vacation autoreply";
+    $self->{is_bounce} = 0;
+    return $self;
+  }
+
+  # Polish auto-reply
+  {
+    last if $message->effective_type eq 'multipart/report';
+    last if !$first_part || $first_part->effective_type ne 'text/plain';
+    my $subject = $message->head->get('Subject');
+    last if !defined($subject);
+    last if $subject !~ /Automatyczna\s+odpowied/;
+    $self->log("looks like a polish autoreply, ignoring.");
+    $self->{type} = "polish autoreply";
+    $self->{is_bounce} = 0;
+    return $self;
+  }
 
   # "Email address changed but your message has been forwarded"
   {
@@ -219,27 +279,6 @@ sub parse {
     return $self;
   }
 
-  # jkburns@compuserve.com has a new CompuServe e-mail address.  The new e-mail
-  # address is johnkingburns@cs.com and CompuServe has automatically forwarded
-  # your message. Please take this opportunity to update your address book with
-  # the new e-mail address.
-  if ($message->effective_type eq "text/plain") {
-    my $string = $message->bodyhandle->as_string;
-    my $forwarded_pos
-      = _match_position($string, qr/automatically.{0,40}forwarded/is);
-
-    my $orig_msg_pos 
-      = _match_position($string, $Returned_Message_Below);
-    if (
-      defined($forwarded_pos)
-      && _position_before($forwarded_pos, $orig_msg_pos)
-    ) {
-      $self->log("message forwarding notification, ignoring");
-      $self->{is_bounce} = 0;
-      return $self;
-    }
-  }
-
   # nonfatal errors usually say they're transient, but sometimes they do it
   # straight out and sometimes it's wrapped in a multipart/report.
   #
@@ -282,8 +321,12 @@ sub parse {
     ) {
       # see MIME::Entity regarding REPLACE
       my $orig_message_id = $orig_message->parts(0)->head->get("message-id");
-      chomp $orig_message_id;
-      $self->log("extracted original message-id $orig_message_id from the original rfc822/message");
+      if ($orig_message_id) {
+        chomp $orig_message_id;
+        $self->log("extracted original message-id $orig_message_id from the original rfc822/message");
+      } else {
+        $self->log("Couldn't extract original message-id from the original rfc822/message");
+      }
       $self->{orig_message_id} = $orig_message_id;
       $self->{orig_message} = $orig_message->parts(0);
     }
@@ -293,14 +336,17 @@ sub parse {
     # message/rfc822.  yow!
 
     if (! $self->{orig_message_id}
-	     and
-	     my ($rfc822_headers) =
+      and
+      my ($rfc822_headers) =
          grep { lc $_->effective_type eq "text/rfc822-headers" } $message->parts
     ) {
       my $orig_head = Mail::Header->new($rfc822_headers->body);
-      chomp ($self->{orig_message_id} = $orig_head->get("message-id"));
-      $self->{orig_header} = $orig_head;
-      $self->log("extracted original message-id $self->{orig_message_id} from text/rfc822-headers");
+      my $message_id = $orig_head->get("message-id");
+      if ($message_id) {
+        chomp ($self->{orig_message_id} = $orig_head->get("message-id"));
+        $self->{orig_header} = $orig_head;
+        $self->log("extracted original message-id $self->{orig_message_id} from text/rfc822-headers");
+      }
     }
   }
 
@@ -315,7 +361,6 @@ sub parse {
     $self->log("couldn't find original message id.");
   }
 
-
   #
   # try to extract email addresses to identify members.
   # we will also try to extract reasons as much as we can.
@@ -340,8 +385,14 @@ sub parse {
     # one line is allowed - it just says "preceded by a blank line".  We very
     # well may put an upper bound on this in the future.
     #
-    # See IIS test in t/.
+    # See t/iis-multiple-bounce.t
     foreach my $para (split /\n{2,}/, $delivery_status_body) {
+
+      # See t/surfcontrol-extra-newline.t - deal with bug #21249
+      $para =~ s/\A\n+//g;
+      # added the following line as part of fix for #41874
+      $para =~ s/\r/ /g;
+
       my $report = Mail::Header->new([split /\n/, $para]);
 
       # Removed a $report->combine here - doesn't seem to work without a tag
@@ -374,11 +425,23 @@ sub parse {
       }
 
       for my $hdr (qw(Reporting-MTA Arrival-Date)) {
-        $report->replace($hdr => $global{$hdr} ||= $report->get($hdr))
+        my $val = $global{$hdr} ||= $report->get($hdr);
+        if (defined($val)) {
+          $report->replace($hdr => $val)
+        }
       }
 
-      next unless my $email = $report->get("original-recipient")
-                           || $report->get("final-recipient");
+      my $email;
+
+      if ($self->{prefer_final_recipient}) {
+        $email = $report->get("final-recipient")
+              || $report->get("original-recipient");
+      } else {
+        $email = $report->get("original-recipient")
+              || $report->get("final-recipient");
+      }
+
+      next unless $email;
 
       # $self->log("email = \"$email\"") if $DEBUG > 3;
 
@@ -386,12 +449,19 @@ sub parse {
       my $reason = $report->get("diagnostic-code");
 
       $email  =~ s/[^;]+;\s*//; # strip leading RFC822; or LOCAL; or system;
-      $reason =~ s/[^;]+;\s*//; # strip leading X-Postfix;
+      if (defined $reason) {
+        $reason =~ s/[^;]+;\s*//; # strip leading X-Postfix;
+      }
 
       $email = _cleanup_email($email);
 
       $report->replace(email      => $email);
-      $report->replace(reason     => $reason);
+      if (defined $reason) {
+        $report->replace(reason     => $reason);
+      } else {
+        $report->delete("reason");
+      }
+
       if (my $status = $report->get('Status')) {
         # RFC 1893... prefer Status: if it exists and is something we know
         # about
@@ -402,6 +472,11 @@ sub parse {
           $report->replace(std_reason => "domain_error");
         } elsif ($status eq "5.2.2") {
           $report->replace(std_reason => "over_quota");
+          # this fits my reading of RFC 3463
+          # FIXME: I suspect there's something wrong with the parsing earlier
+          # that this has to be a regexp rather than a straight comparison
+        } elsif ($status =~ /^5\.4\.4/) {
+          $report->replace(std_reason => "domain_error");
         } else {
           $report->replace(
             std_reason => _std_reason($report->get("diagnostic-code"))
@@ -412,19 +487,35 @@ sub parse {
           std_reason => _std_reason($report->get("diagnostic-code"))
         );
       }
-      my ($host) = $report->get("diagnostic-code") =~ /\bhost\s+(\S+)/;
-      $report->replace( host => ($host)) if $host;
+      my $diag_code = $report->get("diagnostic-code");
 
-      my ($code) = $report->get('diagnostic-code') =~
-         m/ ( ( [245] \d{2} ) \s | \s ( [245] \d{2} ) (?!\.) ) /x;
-
-      $report->replace(smtp_code => $code);
-
-      if (not $report->get("host")) {
-        $report->replace(host => ($report->get("email") =~ /\@(.+)/)[0])
+      my $host;
+      if (defined $diag_code) {
+        ($host) = $diag_code =~ /\bhost\s+(\S+)/;
       }
 
-      if ($report->get("smtp_code") =~ /^2../) {
+      $report->replace(host => ($host)) if $host;
+
+      my ($code);
+
+      if (defined $diag_code) {
+        ($code) = $diag_code =~
+         m/ ( ( [245] \d{2} ) \s | \s ( [245] \d{2} ) (?!\.) ) /x;
+      }
+
+      if ($code) {
+        $report->replace(smtp_code => $code);
+      }
+
+      if (not $report->get("host")) {
+        my $email = $report->get("email");
+        if (defined $email) {
+          my $host = ($email =~ /\@(.+)/)[0];
+          $report->replace(host => $host) if $host;
+        }
+      }
+
+      if ($report->get("smtp_code") and ($report->get("smtp_code") =~ /^2../)) {
         $self->log(
           "smtp code is "
           . $report->get("smtp_code")
@@ -455,30 +546,13 @@ sub parse {
       $self->{is_bounce} = 0;
       return $self;
     }
+
   } elsif ($message->effective_type =~ /multipart/) {
     # but not a multipart/report.  look through each non-message/* section.
+    # See t/corpus/exchange.unknown.msg
 
-    # $self->log("examining non-report multipart...") if $DEBUG > 3;
-
-    # generated by IMS:
-    #
-    # Your message
-    #
-    #  To:      thood@edify.com
-    #  Subject: Red and Blue Online - October 2002
-    #  Sent:    Tue, 15 Oct 2002 13:08:24 -0700
-    #
-    # did not reach the following recipient(s):
-    #
-    # thood@eagle on Tue, 15 Oct 2002 13:18:39 -0700
-    #     The recipient name is not recognized
-    #         The MTS-ID of the original message is: c=us;a=
-    # ;p=edify;l=EAGLE02101520184GF2XLBD
-    #  MSEXCH:IMS:EDIFY:CORP:EAGLE 0 (000C05A6) Unknown Recipient
-    #
-
-    my @delivery_status_parts = grep {
-      $_->effective_type !~ /rfc822|html|image/ and not $_->is_multipart
+    my @delivery_status_parts = grep { $_->effective_type =~ m{text/plain}
+      and not $_->is_multipart
     } $message->parts;
 
     # $self->log("error parts: @{[ map { $_->bodyhandle->as_string }
@@ -486,21 +560,27 @@ sub parse {
 
     push @{$self->{reports}}, $self->_extract_reports(@delivery_status_parts);
 
-  } else {
+  } elsif ($message->effective_type =~ m{text/plain}) {
     # handle plain-text responses
 
+    # This used to just take *any* part, even if the only part wasn't a
+    # text/plain part
+    #
+    # We may have to specifically allow some other types, but in my testing, all
+    # the messages that get here and are actual bounces are text/plain
+    # wby - 20060907
+    
     # they usually say "returned message" somewhere, and we can split on that,
     # above and below.
-
     my $body_string = $message->bodyhandle->as_string || '';
 
     if ($body_string =~ $Returned_Message_Below) {
       my ($stuff_before, $stuff_splitted, $stuff_after) =
-        split $Returned_Message_Below, $message->bodyhandle->as_string, 3;
+        split $Returned_Message_Below, $message->bodyhandle->as_string, 2;
       # $self->log("splitting on \"$stuff_splitted\", " . length($stuff_before)
       # . " vs " . length($stuff_after) . " bytes.") if $DEBUG > 3;
       push @{$self->{reports}}, $self->_extract_reports($stuff_before);
-      $self->{orig_text} = $stuff_before;
+      $self->{orig_text} = $stuff_after;
     } elsif ($body_string =~ /(.+)\n\n(.+?Message-ID:.+)/is) {
       push @{$self->{reports}}, $self->_extract_reports($1);
       $self->{orig_text} = $2;
@@ -555,7 +635,7 @@ sub _extract_reports {
   # we'll keep going that way for each address.
 
   return unless $text;
-  my @split = split(/(\S+\@\S+)/, $text);
+  my @split = split($EMAIL_ADDR_REGEX, $text);
 
   foreach my $i (0 .. $#split) {
     # only interested in the odd numbered elements, which are the email
@@ -565,12 +645,14 @@ sub _extract_reports {
     my $email = _cleanup_email($split[$i]);
 
     if ($split[$i-1] =~ /they are not accepting mail from/) {
-      # aol airmail sender block $self->log("$email is not actually a bouncing
-      # address...") if $DEBUG > 3;
+      # aol airmail sender block
       next;
     }
 
-    # $self->log("looking for the reason that $email bounced...") if $DEBUG > 3;
+    if($split[$i-1] =~ /A message sent by/) {
+      # sender block
+      next;
+    }
 
     my $std_reason = "unknown";
     $std_reason = _std_reason($split[$i+1]) if $#split > $i;
@@ -580,18 +662,20 @@ sub _extract_reports {
     # if we can't figure out the reason, if we're in the delivery-status part,
     # go back up into the text part and try extract_report() on that.
 
-    # $self->log("reason: $std_reason") if $DEBUG > 3;
-
     next if (
       exists $by_email{$email}
       and $by_email{$email}->{std_reason}
       ne "unknown" and $std_reason eq "unknown"
     );
 
+    my $reason = $split[$i-1];
+    $reason =~ s/(.*?). (Your mail to the following recipients could not be delivered)/$2/;
+
     $by_email{$email} = {
       email => $email,
       raw   => join ("", @split[$i-1..$i+1]),
       std_reason => $std_reason,
+      reason => $reason
     };
   }
 
@@ -660,14 +744,20 @@ standardized reasons:
   user_unknown
   over_quota
   domain_error
+  spam
+  message_too_large
   unknown
   no_problemo
 
+The "spam" standard reason indicates that the message bounced because
+the recipient considered it spam.
+
 (no_problemo will only appear if you set {report_non_bounces=>1})
 
-If the bounce message is not structured according to RFC1892, BounceParser will
-still try to return as much information as it can; in particular, you can count
-on "email" and "std_reason" to be present.
+If the bounce message is not structured according to RFC1892,
+BounceParser will still try to return as much information as it can;
+in particular, you can count on "email" and "std_reason" to be
+present.
 
 =cut
 
@@ -758,23 +848,28 @@ RFC1891 and RFC1893.
 
 We understand bounce messages generated by the following MTAs / organizations:
 
- postfix
- sendmail
- qmail
+ Postfix
+ Sendmail
  Exim
- IMS
- Morgan Stanley (ms.com) and emory.edu
+ AOL
+ Yahoo
+ Hotmail
  AOL's AirMail sender-blocking
- Novell Groupwise
+ Microsoft Exchange*
+ Qmail*
+ Novell Groupwise*
+
+ * Items marked with an asterisk currently may return incomplete information.
 
 =head1 SEE ALSO
 
   Used by http://listbox.com/ --- if you like BounceParser and you know it,
   consider Listbox for your mailing list needs!
 
-  Ironically, BounceParser has no mailing list or web site at this time.
+  SVN repository and email list information at:
+  http://emailproject.perl.org/
 
-  See RFC1892, the Multipart/Report Content Type.
+  RFC1892 and RFC1894
 
 =head1 RANDOM OBSERVATION
 
@@ -792,7 +887,7 @@ Ricardo mostly just helped out and managed releases.
 =head1 COPYRIGHT AND LICENSE
 
   Copyright (C) 2003-2006, IC Group, Inc.
-	pobox.com permanent email forwarding with spam filtering
+  pobox.com permanent email forwarding with spam filtering
   listbox.com mailing list services for announcements and discussion
 
 This library is free software; you can redistribute it and/or modify
@@ -808,7 +903,17 @@ it under the same terms as Perl itself.
 sub _std_reason {
   local $_ = shift;
 
-  if (/(domain|host)\s+not\s+found/i) { return "domain_error" }
+  if (!defined $_) {
+    return "unknown";
+  }
+
+  if (/(?:domain|host|service)\s+(?:not\s+found|unknown|not\s+known)/i) {
+    return "domain_error"
+  }
+
+  if (/sorry,\s+that\s+domain\s+isn't\s+in\s+my\s+list\s+of\s+allowed\s+rcpthosts/i) {
+    return "domain_error";
+  }
 
   if (
     /try.again.later/is or
@@ -816,21 +921,27 @@ sub _std_reason {
     /storage/i          or
     /quota/i            or
     /\s552\s/           or
-    /\s#?5\.2\.2\s/                                     # rfc 1893
+    /\s#?5\.2\.2\s/     or                                # rfc 1893
+    /User\s+mailbox\s+exceeds\s+allowed\s+size/i or
+    /Mailbox\s+size\s+limit\s+exceeded/i or
+    /message\s+size\s+\d+\s+exceeds\s+size\s+limit\s+\d+/i or
+    /max\s+message\s+size\s+exceeded/i
   ) {
     return "over_quota";
   }
 
   my $user_re =
    qr'(?: mailbox  | user | recipient | address (?: ee)?
-       | customer | account | e-?mail | <?\S+@\S+>? )'x;
+       | customer | account | e-?mail | <? $EMAIL_ADDR_REGEX >? )'ix;
 
   if (
     /\s \(? \#? 5\.1\.[01] \)? \s/x or                  # rfc 1893
-    /$user_re\s+ (?:\S+\s+)? (?:is\s+)?                 # Generic
-     (?: (?: un|not\s+) known| [dw]oes\s?n[o']?t 
-     (?: exist|found ) | disabled ) /ix or
-    /no\s+(?:such\s+)?$user_re/i or                     # Gmail and other
+    /$user_re\s+(?:\S+\s+)? (?:is\s+)?                  # Generic
+     (?: (?: un|not\s+) (?: known | recognized )
+      | [dw]oes\s?n[o']?t 
+     (?: exist|found ) | disabled | expired ) /ix or
+    /no\s+(?:such)\s+?$user_re/i or                     # Gmail and other (mofified for bug #41874)
+    /unrouteable address/i or                           # bug #41874
     /inactive user/i or                                 # Outblaze
     /unknown local part/i or                            # Exim(?)
     /user\s+doesn't\s+have\s+a/i or                     # Yahoo!
@@ -844,13 +955,22 @@ sub _std_reason {
      (?:suspended|unavailable)/ix or 
     /address is administratively disabled/i or          # Unknown
     /no $user_re\s+(?:here\s+)?by that name/i or        # Unknown
-    /<\S+@\S+> is invalid/i or                          # Unknown
+    /<?$EMAIL_ADDR_REGEX>? is invalid/i or              # Unknown
     /address.*not known here/i or                       # Unknown
     /recipient\s+(?:address\s+)?rejected/i or           # Cox, generic
     /not\s+listed\s+in\s+Domino/i or                    # Domino
     /account not activated/i or                         # usa.net
     /not\s+our\s+customer/i or                          # Comcast
-    /doesn't handle mail for that user/i                # mailfoundry
+    /doesn't handle mail for that user/i or             # mailfoundry
+    /Address\s+does\s+not\s+exist/i or
+    /Recipient\s+<?$EMAIL_ADDR_REGEX>?\s+does\s+not\s+exist/i or
+    /recipient\s+no\s+longer\s+on\s+server/i or # me.com
+    /is\s+not\s+a\s+known\s+user\s+on\s+this\s+system/i or # cam.ac.uk
+    /Rcpt\s+<?$EMAIL_ADDR_REGEX>?\s+does\s+not\s+exist/i or
+    /Mailbox\s+not\s+available/i or
+    /No\s+mailbox\s+found/i or
+    /<?$EMAIL_ADDR_REGEX>?\s+is\s+a\s+deactivated\s+mailbox/i or
+    /Recipient\s+does\s+not\s+exist\s+on\s+this\s+system/i
   ) {
     return "user_unknown";
   }
@@ -860,247 +980,51 @@ sub _std_reason {
     /timed\s+out/i or
     /route\s+to\s+host/i or
     /connection\s+refused/i or
-    /relay/i or
-    /no\s+data\s+record\s+of\s+requested\s+type/i
+    /no\s+data\s+record\s+of\s+requested\s+type/i or
+    /Malformed name server reply/i or
+    /as\s+a\s+relay,\s+but\s+I\s+have\s+not\s+been\s+configured\s+to\s+let/i or
+    /550\s+relay\s+not\s+permitted/i or
+    /550\s+relaying\s+denied/i or
+    /Relay\s+access\s+denied/i or
+    /Relaying\s+denied/i or
+    /No\s+such\s+domain\s+at\s+this\s+location/i
   ) {
     return "domain_error";
   }
 
+  if (
+    /Blocked\s+by\s+SpamAssassin/i or
+    /spam\s+rejection/i or
+    /identified\s+SPAM,\s+message\s+permanently\s+rejected/i or
+    /Mail\s+appears\s+to\s+be\s+unsolicited/i or
+    /Message\s+rejected\s+as\s+spam\s+by\s+Content\s+Filtering/i or
+    /message\s+looks\s+like\s+SPAM\s+to\s+me/i or
+    /your\s+message\s+has\s+triggered\s+a\s+SPAM\s+block/i or
+    /Spam\s+detected/i or
+    /Message\s+looks\s+like\s+spam/i
+  ) {
+    return "spam";
+  }
+
+  if (
+    /RESOLVER.RST.RecipSizeLimit/i
+  ) {
+    return "message_too_large";
+  }
+
   return "unknown";
 }
 
 # ---------------------------------------------------------------------
-# 		       preprocessors
+# preprocessors
 # ---------------------------------------------------------------------
 
-sub p_ms {
-  my $self    = shift;
-  my $message = shift;
-  my $domain = qr(ms\.com|emory\.edu|wharton\.upenn\.edu|midway\.uchicago\.edu)i;
-
-  # From: Mail Delivery Subsystem <MAILER-DAEMON@ms.com>
-  # Subject: Undeliverable Mail
-  # X-Listbox-Reason: samantha.wright@msdw.com user unknown
-  # Lines: 116
-  #
-  #    ----- The following addresses had permanent fatal errors -----
-  #
-  # wrights: User unknown
-  #
-  #    ----- Original message follows -----
-  #
-  # Received: from hqvsbh1.ms.com (hqvsbh1-i0.morgan.com [199.89.99.101])
-  #         by pismh4.ms.com (8.8.5/imap+ldap v2.4) with ESMTP id QAA23003
-  #         for <wrights@ms.com>; Thu, 19 Sep 2002 16:52:55 -0400 (EDT)
-  # Received: from hqvsbh1.ms.com (localhost [127.0.0.1])
-  #         by localhost.ms.com (Postfix) with SMTP id 35090205D5
-  #         for <samantha.wright@msdw.com>; Thu, 19 Sep 2002 16:52:55 -0400 (EDT)
-
-  # -------------------------------------------------- case 2
-
-  # 20030218-02:00:53 root@frodo:~listbox/Archive# less afe-chicago@v2.listbox.com/200209/18/1032373917.29028_1.frodo
-  # Return-Path: <>
-  # Delivered-To: listbox+trampoline+282+137642+ec84f782@frodo.listbox.com
-  # Received: from juliet.cc.emory.edu (juliet.cc.emory.edu [170.140.204.2])
-  #         by frodo.listbox.com (Postfix) with ESMTP id E9BF78038
-  #         for <listbox+trampoline+282+137642+ec84f782@v2.listbox.com>; Wed, 18 Sep 2002 12:41:29 -0400 (EDT)
-  # Received: (from root@localhost)
-  #         by juliet.cc.emory.edu (8.10.2/8.10.2) with X.500 id g8IGfTN03537
-  #         for listbox+trampoline+282+137642+ec84f782@v2.listbox.com; Wed, 18 Sep 2002 12:41:29 -0400 (EDT)
-  # Date: Wed, 18 Sep 2002 12:41:29 -0400 (EDT)
-  # Message-Id: <200209181641.g8IGfTN03537@juliet.cc.emory.edu>
-  # To: listbox+trampoline+282+137642+ec84f782@v2.listbox.com
-  # From: MAILER-DAEMON@emory.edu
-  # Subject: undeliverable mail for rchiri
-  # X-Listbox-Reason: rchiri@emory.edu user unknown
-  # Lines: 68
-  #
-  # The following errors occurred when trying to deliver the attached mail:
-  #
-  # rchiri: User unknown
-  #
-  # ------- The original message sent:
-  #
-  # Received: from frodo.listbox.com (frodo.listbox.com [208.210.125.58])
-  #         by juliet.cc.emory.edu (8.10.2/8.10.2) with ESMTP id g8IGfT003531
-  #         for <rchiri@emory.edu>; Wed, 18 Sep 2002 12:41:29 -0400 (EDT)
-  # Received: by frodo.listbox.com (Postfix, from userid 1003)
-  #         id C5DBE8036; Wed, 18 Sep 2002 12:41:28 -0400 (EDT)
-
-  # ------------------------------------------------------------ case 3
-  #  afe-chicago@v2.listbox.com/200209/18/1032373971.29028_1.frodo
-  #  afe-chicago@v2.listbox.com/200209/18/1032373938.29028_1.frodo
-
-  # From: MAILER-DAEMON@wharton.upenn.edu
-  # Message-Id: <200209181642.MAA14592@barter.wharton.upenn.edu>
-  # To: listbox+trampoline+282+138095+03cc3f4e@v2.listbox.com
-  # Subject: Returned mail - nameserver error report
-  # X-Listbox-Reason: weilong.ye.wa98@wharton.upenn.edu unknown
-  # Lines: 79
-  #
-  #  --------Message not delivered to the following:
-  #
-  #  weilong.ye.wa98    No matches to nameserver query
-  #
-  #  --------Error Detail (phquery V4.4):
-  #
-  #  The message, "No matches to nameserver query," is generated whenever
-  #  the ph nameserver fails to locate either a ph alias or name field that
-  #  matches the supplied name.  The usual causes are typographical errors or
-  #  the use of nicknames.  Recommended action is to use the ph program to
-  #  determine the correct ph alias for the individuals addressed.  If ph is
-  #  not available, try sending to the most explicit form of the name, e.g.,
-  #  if mike-fox fails, try michael-fox or michael-j-fox.
-  #
-  #
-  #  --------Unsent Message below:
-  #
-  # Received: from frodo.listbox.com (frodo.listbox.com [208.210.125.58]) by barter.wharton.upenn.edu with ESMTP (8.9.3 (PHNE_18546)/8.7
-  # .1) id MAA14570 for <weilong.ye.wa98@wharton.upenn.edu>; Wed, 18 Sep 2002 12:42:35 -0400 (EDT)
-  # Received: by frodo.listbox.com (Postfix, from userid 1003)
-  #
-
-  # $self->log("p_ms: didn't match domain") and
-  return
-    unless ($message->head->get("from")||'') =~ /MAILER-DAEMON\@($domain)\b/i;
-
-  $domain = $1;
-
-  # $self->log("p_ms: couldn't find \"--- original message ---\" separator") and
-  return unless not $message->is_multipart and
-  my ($error_part, $orig_message) = split(
-    /.*-----.*(?:Original message follows|The original message sent:|unsent message below).*/i,
-    $message->bodyhandle->as_string
-  );
-
-  # $self->log("p_ms: couldn't find \"following message ... errors") and
-  return unless $error_part =~ /The following addresses had permanent fatal errors|The following errors occurred when trying to deliver the attached mail:|message not delivered/i;
-
-  # $self->log("p_ms: couldn't find Received: header in orig message.") and
-  return unless $orig_message =~ /^Received: from/m;
-
-  my @error_lines;
-  if ($error_part =~ /error detail.*phquery/i) {
-    @error_lines = (
-      grep { /\S/ }
-      grep { ! /error detail/i }
-      grep { /message.*not delivered to the following/i .. /error detail/i }
-      split /\n/, $error_part
-    );
-  } else {
-    @error_lines = grep {
-      ! /addresses had permanent fatal errors/
-      &&
-      /The following errors occurred when trying to deliver the attached mail:/
-			&&
-			/\S/
-    } split /\n/, $error_part;
-  }
-
-  my @new_errors;
-  foreach my $error_line (@error_lines) {
-    push @new_errors, $error_line
-      and next
-      unless my ($address, $error) = $error_line =~ /^\s*(\S+) (.*)/;
-    # $self->log("considering address $address error $error") if $DEBUG > 3;
-    if ($address !~ /\@/) {
-      $address = _cleanup_email($address);
-      if ($orig_message =~ /for\s+<(\Q$address\E\@\S+)>/i) {
-        push @new_errors, "$1: $error";
-        push @new_errors, "$address\@$domain: $error";
-        $self->log("p_ms: rewrote $address to $address\@$domain ($error)");
-      }
-    }
-  }
-  # $self->log("p_ms: rewrote message. new errors: @new_errors") if $DEBUG > 3;
-
-  return $self->new_plain_report(
-    $message,
-    join ("\n", @new_errors),
-    $orig_message
-  );
-}
-
-sub p_compuserve {
-  my ($self, $message) = @_;
-
-  # afe-chicago@v2.listbox.com/200209/18/1032373866.29028_0.frodo
-  # From: CompuServe Postmaster <postmaster@compuserve.com>
-  # Subject: Undeliverable Message
-  # Sender: CompuServe Postmaster <auto.reply@compuserve.com>
-  # To: listbox+trampoline+282+137177+366d45b9@v2.listbox.com
-  # X-Listbox-Reason: lallison@compuserve.com user unknown
-  # Lines: 81
-  #
-  # Receiver not found: lallison
-  #
-  #
-  # Your message could not be delivered as addressed.
-  #
-  # --- Message From Postmaster ---
-  #
-  # Subject: Addressing CompuServe Mail users
-  #
-  # Please contact postmaster@compuserve.com if you need additional formatting
-  # information for other types of addresses.
-  #
-  # Cordially,
-  #
-  # The Electronic Postmaster
-  #
-  # --- Returned Message ---
-  #
-  # Sender: listbox+trampoline+282+137177+366d45b9@v2.listbox.com
-  # Received: from frodo.listbox.com (frodo.listbox.com [208.210.125.58])
-  #         by siaag1af.compuserve.com (8.9.3/8.9.3/SUN-1.14) with ESMTP id MAA16547
-  #         for <lallison@compuserve.com>; Wed, 18 Sep 2002 12:40:17 -0400 (EDT)
-  # Received: by frodo.listbox.com (Postfix, from userid 1003)
-  #         id 65A1F8056; Wed, 18 Sep 2002 12:40:16 -0400 (EDT)
-  #
-  return if not $message->head->get("from") =~ /\@compuserve\.com/i;
-  return if $message->is_multipart;
-  return unless $message->bodyhandle->as_string =~ /Receiver not found:/;
-
-  my ($stuff_before, $stuff_after)
-    = split(/.*Returned Message.*/i, $message->bodyhandle->as_string);
-
-  $stuff_before =~ s/Your message could not be delivered as addressed.*//is;
-
-  my @new_errors;
-  for (split /\n/, $stuff_before) {
-    if (my ($receiver) = /Receiver not found:\s*(\S+)/) {
-      if ($receiver !~ /\@/) {
-        push @new_errors, "$receiver\@compuserve\.com: Receiver not found";
-        next;
-      }
-    }
-    push @new_errors, $_;
-  }
-  return $self->new_plain_report(
-    $message,
-    join ("\n", @new_errors),
-    $stuff_after
-  );
-}
-
 sub p_ims {
   my $self    = shift;
   my $message = shift;
 
-  # Your message
-  #
-  #   To:      slpark@msx.ndc.mc.uci.edu
-  #   Subject: Penn Women Paving The Way - Register Now!
-  #   Sent:    Thu, 19 Sep 2002 13:40:20 -0700
-  #
-  # did not reach the following recipient(s):
-  #
-  # c=US;a= ;p=NDC;o=ORANGE;dda:SMTP=slpark@msx.ndc.mc.uci.edu; on Thu, 19 Sep
-  # 2002 13:53:00 -0700
-  #     The recipient name is not recognized
-  #         The MTS-ID of the original message is: c=us;a=
-  # ;p=ndc;l=LEA0209192052TAM7PVWM
-  #     MSEXCH:IMS:NDC:ORANGE:LEA 0 (000C05A6) Unknown Recipient
+  # Mangle Exchange messages into a format we like better
+  # see t/corpus/exchange.unknown.msg
 
   return
     unless ($message->head->get("X-Mailer")||'') =~ /Internet Mail Service/i;
@@ -1117,40 +1041,8 @@ sub p_ims {
       $io->print($actual_error);
       $io->close;
     }
-    # $self->log("rewrote IMS error text to " . $error_part->as_string) if
-    # $DEBUG > 3;
+
   } else {
-    # X-Mailer: Internet Mail Service (5.5.2654.52)
-    # X-MS-Embedded-Report:
-    # MIME-Version: 1.0
-    # X-WSS-ID: 11966AB2117930-01-01
-    # Content-Type: text/plain
-    # Content-Transfer-Encoding: 7bit
-    # X-Listbox-Reason: jfrancl@sidley.com user unknown
-    # Lines: 66
-    #
-    # Your message
-    #
-    #   To:      USER_EMAIL@frodo.listbox.com
-    #   Subject: You have an online postcard waiting for you!
-    #   Sent:    Wed, 18 Sep 2002 11:00:36 -0500
-    #
-    # did not reach the following recipient(s):
-    #
-    # jfrancl@sidley.com on Wed, 18 Sep 2002 11:40:15 -0500
-    #     The recipient name is not recognized
-    #         The MTS-ID of the original message is: c=us;a= ;p=sidley
-    # austin;l=CHEXCHANGE10209181640TFHKWW8X
-    #     MSEXCH:IMS:Sidley & Austin:Chicago:CHEXCHANGE1 0 (000C05A6) Unknown
-    # Recipient
-    #
-    #
-    #
-    # -----
-    # Message-ID: <E17rhFk-00085B-00@erie.vervehosting.com>
-    # From: Bonnie Eisner <owner-afe-chicago@v2.listbox.com>
-    # To: USER_EMAIL@frodo.listbox.com
-    #
 
     return unless my ($actual_error)
       = $message->bodyhandle->as_string
@@ -1213,33 +1105,16 @@ sub p_aol_senderblock {
   return $message;
 }
 
-sub p_novell_groupwise_5_2 {
+sub p_novell_groupwise {
+
+  # renamed from p_novell_groupwise_5_2 - hopefully we can deal with most / all
+  # versions and create test cases / fixes when we can't
+  #
+  # See t/various-unknown.t and t/corpus/novell-*.msg for some recent examples.
+
   my $self    = shift;
   my $message = shift;
 
-  # X-Mailer: Novell GroupWise 5.2
-  # Date: Thu, 19 Sep 2002 13:52:48 -0700
-  # From: Mailer-Daemon@chrismill.com
-  # To: listbox+trampoline+260+123494+341495fc@v2.listbox.com
-  # Subject: Message status - undeliverable
-  # Mime-Version: 1.0
-  # Content-Type: multipart/mixed; boundary="=_D68A6C30.E687E8F9"
-  # X-Listbox-Reason: lsinger@chrismill.com user unknown
-  # Lines: 57
-  #
-  # --=_D68A6C30.E687E8F9
-  # Content-Type: text/plain; charset=US-ASCII
-  # Content-Disposition: inline
-  #
-  # The message that you sent was undeliverable to the following:
-  #         lsinger (user not found)
-  #
-  # Possibly truncated original message follows:
-  #
-  # --=_D68A6C30.E687E8F9
-  # Content-Type: message/rfc822
-  #
-
   return unless ($message->head->get("X-Mailer")||'') =~ /Novell Groupwise/i;
   return unless $message->effective_type eq "multipart/mixed";
   return unless my ($error_part)
@@ -1247,6 +1122,9 @@ sub p_novell_groupwise_5_2 {
 
   my ($host) = $message->head->get("From") =~ /\@(\S+)>?/;
 
+  # A lot of times, Novell returns just the LHS; this makes it difficult /
+  # impossible in many cases to guess the recipient address. MBP makes an
+  # attempt here.
   my @new_output;
   for (split /\n/, $error_part->bodyhandle->as_string) {
     if (/^(\s*)(\S+)(\s+\(.*\))$/) {
@@ -1262,7 +1140,7 @@ sub p_novell_groupwise_5_2 {
   }
 
   push @new_output,
-    ("# rewritten by BounceParser: p_novell_groupwise_5_2()", "");
+    ("# rewritten by BounceParser: p_novell_groupwise()", "");
 
   if (my $io = $error_part->open("w")) {
     $io->print(join "\n", @new_output);
@@ -1357,6 +1235,16 @@ sub _construct_delivery_status_paras {
   return @new_output;
 }
 
+sub _construct_diagnostic_code {
+  my %by_email = %{shift()};
+  my $email = shift;
+  join (" ",
+  "Diagnostic-Code: X-BounceParser;",
+  ($by_email{$email}->{'host'} ? "host $by_email{$email}->{'host'} said:" : ()),
+  ($by_email{$email}->{'smtp_code'}),
+  (join ", ", @{ $by_email{$email}->{'errors'} }));
+}
+
 sub _analyze_smtp_transcripts {
   my $self = shift;
   my $plain_smtp_transcript = shift;
@@ -1365,9 +1253,12 @@ sub _analyze_smtp_transcripts {
 
   # parse the text part for the actual SMTP transcript
   for (split /\n\n|(?=>>>)/, $plain_smtp_transcript) {
-    # $self->log("_analyze_smtp_transcripts: $_") if $DEBUG > 3;
-
     $email = _cleanup_email($1) if /RCPT TO:\s*(\S+)/im;
+
+    if (/The\s+following\s+addresses\s+had\s+permanent\s+fatal\s+errors\s+-----\s+\<(.*)\>/im) {
+      $email = _cleanup_email($1);
+    }
+
     $by_email{$email}->{host} = $host if $email;
 
     if (/while talking to (\S+)/im) {
