@@ -1,9 +1,13 @@
 package Mail::DeliveryStatus::Report;
 
+our $VERSION = '1.525';
+$VERSION = eval $VERSION;
+
 use Mail::Header;
 use strict;
+use warnings;
 use vars qw(@ISA);
-@ISA = qw(Mail::Header);
+BEGIN { @ISA = qw(Mail::Header) };
 
 # i just don't like how Mail::Header leaves a \n at the end of everything
 # meng
